package edu.tum.motras.algorithms;

import static edu.tum.motras.algorithms.util.Util.checkArguments;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.ImmutableSet;

import edu.tum.motras.algorithms.util.StatePair;
import edu.tum.xmts.Action;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;

public abstract class AbstractComposition extends WaitingQueueAlgorithm<StatePair> implements Algorithm<State> {

	protected Set<String> synchronizingAlphabet;

	public AbstractComposition() {
		synchronizingAlphabet = Collections.emptySet();
	}

	public AbstractComposition(String synchronizingAlphabet) {
		setConfiguration(synchronizingAlphabet);
	}

	@Override
	public boolean isConfigurable() {
		return true;
	}

	@Override
	public void setConfiguration(String conf) {
		this.synchronizingAlphabet = ImmutableSet.copyOf(conf.replaceAll(" ", "").split(","));
	}

	private static boolean isEqual(Action action1, Action action2) {
		return action1.getId().equals(action2.getId());
	}

	@Override
	public State apply(State... input) {
		checkArguments(this, input);

		run(new StatePair(input[0], input[1]));
		State initialState = transitionSystem.getInitialState();
		transitionSystem = null;
		return initialState;
	}

	@Override
	public int getArity() {
		return 2;
	}

	public Set<String> getSynchronizingAlphabet() {
		return synchronizingAlphabet;
	}

	private State getState(State state1, State state2) {
		StatePair pair = new StatePair(state1, state2);
		queue(pair);
		return stateMapper.map(pair);
	}

	protected Transition getTransition(State context, Action action, State state) {
		Transition transition = context.getTransition(actionMapper.map(action), state);

		if (transition == null) {
			transition = context.createTransition(actionMapper.map(action), state);
		}

		return transition;
	}

	protected boolean isSynchronized(Action action) {
		return synchronizingAlphabet.contains(action.getId());
	}

	protected abstract void setObligation(State state, State leftState, State rightState,
			Map<Transition, Transition> translationMap);

	@Override
	protected void process(StatePair statePair) {
		State state = stateMapper.map(statePair);
		Map<Transition, Transition> translationMap = new HashMap<>();

		/* May */
		for (Transition transition1 : statePair.get(1).getTransitions()) {
			boolean isSynchronized = isSynchronized(transition1.getAction());

			if (!isSynchronized) {
				State nextState = getState(transition1.getState(), statePair.get(2));
				Transition newTransition = getTransition(state, transition1.getAction(), nextState);
				translationMap.put(transition1, newTransition);
				continue;
			}

			for (Transition transition2 : statePair.get(2).getTransitions()) {
				if (isEqual(transition1.getAction(), transition2.getAction())) {

					State nextState = getState(transition1.getState(), transition2.getState());
					Transition newTransition = getTransition(state, transition1.getAction(), nextState);
					translationMap.put(transition1, newTransition);
					translationMap.put(transition2, newTransition);
				}
			}
		}

		for (Transition transition2 : statePair.get(2).getTransitions()) {
			boolean isSynchronized = isSynchronized(transition2.getAction());

			if (!isSynchronized) {
				State nextState = getState(statePair.get(1), transition2.getState());
				Transition newTransition = getTransition(state, transition2.getAction(), nextState);
				translationMap.put(transition2, newTransition);
			}
		}

		setObligation(state, statePair.get(1), statePair.get(2), translationMap);
	}

}