package edu.tum.motras.algorithms;

import static edu.tum.motras.algorithms.util.Util.checkArguments;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import com.google.common.collect.ImmutableSet;

import edu.tum.motras.algorithms.util.StateSet;
import edu.tum.xmts.Action;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;

public class DeterministicHullBMTS extends WaitingQueueAlgorithm<StateSet> implements Algorithm<State> {

	@Override
	public State apply(State... input) {
		checkArguments(this, input);
		run(new StateSet(Collections.singleton(input[0])));
		return transitionSystem.getInitialState();
	}

	@Override
	public int getArity() {
		return 1;
	}

	@Override
	public Set<Type> getSupportedTypes() {
		return ImmutableSet.of(Type.MTS, Type.DMTS, Type.BMTS);
	}

	@Override
	public String toString() {
		return "Deterministic Hull (BMTS)";
	}

	private State getState(Set<State> states) {
		StateSet set = new StateSet(states);
		queue(set);
		return stateMapper.map(set);
	}

	@Override
	protected void process(StateSet element) {
		State state = stateMapper.map(element);

		for (Action action : element.getActions()) {
			Set<State> successors = element.getSuccessors(action);

			if (successors.isEmpty()) {
				continue;
			}

			State nextState = getState(successors);
			state.createTransition(actionMapper.map(action), nextState);
		}

		Set<Set<Transition>> tran = new HashSet<>();

		for (State orginalState : element.getStates()) {
			for (Set<Transition> originalSetOfTransitions : orginalState.getObligation().getTran()) {
				// What about actions only occurring in a single state.

				Set<Transition> newTransitions = new HashSet<>();

				for (Transition transition : originalSetOfTransitions) {
					Transition newTransition = state.getTransitions(actionMapper.map(transition.getAction()))
							.iterator().next();
					newTransitions.add(newTransition);
				}

				tran.add(newTransitions);
			}
		}

		state.setObligation(tran);
	}

	@Override
	protected void setup(TransitionSystem newSystem) {
		newSystem.setId("deterministicHull");
		newSystem.setType(Type.BMTS);
	}

	@Override
	public boolean isConfigurable() {
		return false;
	}

	@Override
	public void setConfiguration(String conf) {
		throw new UnsupportedOperationException("This algorithm cannot be configured");
	}
}
