package edu.tum.motras.algorithms;

import static edu.tum.motras.algorithms.util.Util.checkArguments;

import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.eclipse.jdt.annotation.NonNullByDefault;

import com.google.common.collect.ImmutableSet;

import edu.tum.xmts.Parameter;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.formula.RemoveVisitor;
import edu.tum.xmts.formula.SimplifyVisitor;
import edu.tum.xmts.formula.TranslationVisitor;

public class RandomRefinement extends WaitingQueueAlgorithm<State> implements Algorithm<State> {

	@Override
	public State apply(State... input) throws IllegalArgumentException {
		checkArguments(this, input);

		run(input[0]);
		return transitionSystem.getInitialState();
	}

	@Override
	public int getArity() {
		return 1;
	}

	@Override
	public Set<Type> getSupportedTypes() {
		return ImmutableSet.of(Type.MTS, Type.DMTS, Type.BMTS, Type.PMTS);
	}

	@Override
	public String toString() {
		return "Random Refinement";
	}

	private State getState(State state) {
		State newState = stateMapper.map(state);
		queue(state);
		return newState;
	}

	@NonNullByDefault
	@Override
	protected void process(State element) {
		State state = stateMapper.map(element);

		Map<Transition, Transition> transitionMap = new IdentityHashMap<>(element.getTransitions().size());

		for (Transition transition : element.getTransitions()) {
			State nextState = getState(transition.getState());
			Transition newTransition = state.createTransition(actionMapper.map(transition.getAction()), nextState);
			transitionMap.put(transition, newTransition);
		}

		Formula oldFormula = element.getObligation().getFormula();

		for (Parameter parameter : element.getContext().parameters.getElements()) {
			oldFormula = oldFormula.accept(new RemoveVisitor(parameter));
		}

		@SuppressWarnings("null")
		TranslationVisitor<Transition> translationVisitor = new TranslationVisitor<>(transitionMap);
		oldFormula = oldFormula.accept(new SimplifyVisitor());
		oldFormula = oldFormula.accept(translationVisitor);
		oldFormula = oldFormula.and(new Formula.Atom<>(state.getTransitions().get(
				new Random().nextInt(state.getTransitions().size()))));

		state.setObligation(oldFormula);
	}

	@Override
	protected void setup(TransitionSystem transitionSystem) {
		transitionSystem.setId("RandomRefinement");
		transitionSystem.setType(Type.BMTS);
	}

	@Override
	public boolean isConfigurable() {
		return false;
	}

	@Override
	public void setConfiguration(String conf) {
		throw new UnsupportedOperationException("This algorithm cannot be configured");
	}
}
