package edu.tum.motras.algorithms.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Arrays;
import java.util.List;

/**
 * 
 * @author Salomon Sickert
 */
public abstract class ExternalTool<Input, Output> {

	private final ProcessBuilder processBuilder;

	protected ExternalTool(String executable, List<String> parameters) throws IOException {
		File baseDir = new File(System.getProperty("user.dir"));
		File libDir = null;
		File logDir = null;

		while (baseDir.getParent() != null) {
			libDir = new File(baseDir, "lib");

			if (libDir.isDirectory()) {
				break;
			}

			baseDir = baseDir.getParentFile();
		}

		if (libDir == null) {
			throw new IOException("Unable to find library directory.");
		}

		logDir = new File(baseDir, "log");

		if (!logDir.exists()) {
			logDir.mkdir();
		}

		processBuilder = new ProcessBuilder(new File(libDir, executable).getAbsolutePath());
		processBuilder.directory(libDir);
		processBuilder.command().addAll(parameters);
		processBuilder.redirectError(new File(logDir, executable + ".log"));
	}

	protected ExternalTool(String executable, String... parameters) throws IOException {
		this(executable, Arrays.asList(parameters));
	}

	protected void addParameter(String parameter) {
		processBuilder.command().add(parameter);
	}

	private static boolean isRunning(Process process) {
		try {
			process.exitValue();
			return false;
		} catch (IllegalThreadStateException exp) {
			return true;
		}
	}

	private long lastExecutionTimeMs = -1;

	public Output run(Input input) throws IOException {
		StringWriter buffer = new StringWriter();
		Process process = processBuilder.start();
		int exitValue = -1;

		long time = System.currentTimeMillis();

		try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(process.getOutputStream()))) {
			writeInput(writer, input);
		}

		try (BufferedReader reader = new BufferedReader(new InputStreamReader(process.getInputStream()))) {
			String line;

			do {
				try {
					while ((line = reader.readLine()) != null) {
						buffer.write(line);
					}

					process.waitFor();
				} catch (InterruptedException e) {
					// NOP
				}
			} while (isRunning(process));

			exitValue = process.exitValue();
		} finally {
			process.destroy();
		}

		lastExecutionTimeMs = System.currentTimeMillis() - time;

		return readOutput(exitValue, buffer.toString());
	}

	public long getLastExecutionTime() {
		return lastExecutionTimeMs;
	}

	protected abstract Output readOutput(int exitValue, String output);

	protected abstract void writeInput(Writer writer, Input input) throws IOException;
}
