package edu.tum.motras.benchmarker;

import java.io.File;
import java.io.IOException;
import java.util.Random;
import java.util.concurrent.TimeoutException;

import edu.tum.motras.algorithms.Algorithm;
import edu.tum.motras.algorithms.ModalRefinementQSat;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.utils.XMTSGenerator;
import edu.tum.xmts.utils.XMTSGenerator.XMTSGeneratorConfiguration;

public class BenchmarkSuiteGenerator {

	private static void w(String s) {
		System.out.print(s);
	}

	public static void main(String[] args) throws IOException {
		// Configuration for the benchmark suite.

		// Number of pairs for for true and false
		final int limit = 5;

		// Changes applied to the original system
		final int[] distance = { 0, 0, 1, 1 };

		// Algorithm for which the benchmark suite is generated
		final Algorithm<Boolean> algorithm = new ModalRefinementQSat();

		// Dimensions of the benchmark suite
		final int[] alphabetSteps = { 2 };
		final int[] degreeSteps = { 2, 5, 10 };
		final int[] sizeSteps = { 50, 100, 150, 200, 250, 300 };
		final Type[] typeSteps = { Type.MTS, Type.PMTS, Type.PMTS, Type.PMTS, Type.PMTS };
		final int[] parameterSteps = { 0, 1, 5, 20 };

		int i = 0;

		w("Algorithm: " + algorithm.getClass().getSimpleName() + "\n");

		for (int alphabetSize : alphabetSteps) {
			w("* Alphabet: " + alphabetSize + "\n");

			for (int degree : degreeSteps) {
				w("** Degree: " + degree + "\n");

				/* Print Header */
				w("\t\t");
				for (int size : sizeSteps) {
					w(size + "\t");
				}
				w("\n");

				/* Column */
				int index = -1;
				for (Type type : typeSteps) {
					if (type == Type.PMTS) {
						index++;
					}

					w(type.name() + (index == -1 ? "\t" : " (" + parameterSteps[index] + ")") + "\t");
					for (int size : sizeSteps) {
						SimplifiedXMTSGeneratorConfiguration configuration = new SimplifiedXMTSGeneratorConfiguration(
								type, index == -1 ? 0 : parameterSteps[index], alphabetSize, size, degree);

						Benchmark benchmark = generateSystems(configuration, distance, limit, algorithm);

						w(".\t");

						File benchmarkFolder = new File("benchmark_" + i);
						benchmarkFolder.mkdir();
						benchmark.writeTo(benchmarkFolder);
						i++;
					}
					w("\n");
				}
				w("\n");

			}
			w("\n");
		}

	}

	private static XMTSGeneratorConfiguration obtainConfiguration(SimplifiedXMTSGeneratorConfiguration parameterObject) {
		XMTSGeneratorConfiguration configuration = new XMTSGeneratorConfiguration();

		/* Copy Configuration */
		configuration.type = parameterObject.type;
		configuration.actions = parameterObject.alphabetSize;
		configuration.systemNodeDegree = parameterObject.degree;
		configuration.systemSize = parameterObject.states;
		configuration.parameters = (parameterObject.type == Type.PMTS ? parameterObject.parameters : 0);

		/* Disable Organic Structure */
		configuration.clusterInterfaceNodes = 1;
		configuration.clusterNodeDegree = 0;
		configuration.clusterSize = 1;

		configuration.deterministic = false;

		return configuration;
	}

	private static Benchmark generateSystems(SimplifiedXMTSGeneratorConfiguration parameterObject, int[] distance,
			int limit, Algorithm<Boolean> checker) {
		final XMTSGenerator generator = new XMTSGenerator();
		final XMTSGeneratorConfiguration configuration = obtainConfiguration(parameterObject);

		final Benchmark set = new Benchmark(parameterObject, distance, limit);
		final Random random = new Random(42l);
		final TransitionSystemModifier modifier = new TransitionSystemModifier();

		int i;

		for (i = 0; i < 200 && !set.isComplete(); i++) {
			long seed = Math.abs(random.nextLong());

			Benchmark.Pair pair;

			TransitionSystem system = generator.generate(configuration, seed, seed + 1);
			TransitionSystem system2 = modifier.preformChanges(system, distance[0], distance[1], distance[2],
					distance[3]);

			if (system2.getInitialState() == null) {
				system2.setInitialState(system2.states.getElements().get(0));
			}

			try {
				pair = new Benchmark.Pair(system, system2);

				if (checker.apply(system.getInitialState(), system2.getInitialState())) {
					set.offerTrue(pair);
				} else {
					set.offerFalse(pair);
				}

				pair = new Benchmark.Pair(system2, system);

				if (checker.apply(system2.getInitialState(), system.getInitialState())) {
					set.offerTrue(pair);
				} else {
					set.offerFalse(pair);
				}
			} catch (IOException | IllegalArgumentException | TimeoutException e) {
				System.err.println("Exception: " + e.getMessage());
				e.printStackTrace();
				System.err.println("Skipping pair.");
			}
		}

		if (i == 100) {
			System.err.print("Tried in vain. Could not generate enough pairs.");
		}

		return set;
	}
}
