package edu.tum.motras.benchmarker;

import java.util.NoSuchElementException;
import java.util.Random;

import edu.tum.xmts.Action;
import edu.tum.xmts.Obligation;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.formula.CheckContextVisitor;
import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.formula.RemoveVisitor;
import edu.tum.xmts.formula.ReplaceVisitor;
import edu.tum.xmts.formula.SimplifyVisitor;
import edu.tum.xmts.utils.RandomUtil;

public class TransitionSystemModifier {

	private final Random randomness = new Random(4);

	public TransitionSystem preformChanges(TransitionSystem original, int actionDistance, int stateDistance,
			int transitionDistance, int obligationDistance) {

		TransitionSystem current = original.clone();

		current.setId(current.getId() + "_modified");

		// Action insertion and deletion
		for (int i = 0; i < actionDistance; i++) {
			if (randomness.nextBoolean()) {
				current.actions.create();
			} else {
				Action action = selectRandomAction(current);
				current.actions.remove(action);
			}
		}

		// State insertion and deletion
		for (int i = 0; i < stateDistance; i++) {
			if (randomness.nextBoolean()) {
				current.states.create();
			} else {
				State state = selectRandomState(current);

				if (current.getInitialState() != state) {
					current.states.remove(state);
				} else {
					i--;
				}
			}
		}

		// Transition insertion, deletion and label change
		for (int i = 0; i < transitionDistance; i++) {

			try {
				State src, dst;

				Action a;
				Transition t, t2;

				switch (randomness.nextInt(3)) {
				case 0:
					// Transition insertion
					src = selectRandomState(current);
					dst = selectRandomState(current);
					a = selectRandomAction(current);

					if (null == src.getTransition(a, dst)) {
						src.createTransition(a, dst);
					} else {
						i--;
					}

					break;

				case 1:
					// Transition deletion
					t = selectRandomTransition(current);
					t.getContext().removeTransition(t);
					break;

				case 2:
					// Transition label change
					src = selectRandomState(current);
					a = selectRandomAction(current);
					t = selectRandomTransition(src);

					if (null == src.getTransition(a, t.getState())) {
						t2 = src.createTransition(a, t.getState());
					} else {
						i--;
						break;
					}

					ReplaceVisitor visitor = new ReplaceVisitor(t, t2);
					Formula f = src.getObligation().getFormula().accept(visitor);

					src.setObligation(f);
					src.removeTransition(t);

					CheckContextVisitor visitor2 = new CheckContextVisitor(src);
					src.getObligation().getFormula().accept(visitor2);
					break;

				default:
					throw new IllegalStateException();
				}
			} catch (NoSuchElementException e) {
				i--;
			}
		}

		for (int i = 0; i < obligationDistance; i++) {
			State state = selectRandomState(current);
			Obligation obligation = state.getObligation();
			SimplifyVisitor simplifyVisitor = new SimplifyVisitor();

			try {
				if (randomness.nextBoolean()) {
					// Add t to the obligation
					Transition t = selectRandomTransition(state);
					Formula f = obligation.getFormula().and(new Formula.Atom<>(t));
					state.setObligation(f.accept(simplifyVisitor));
				} else {
					// Remove t from the obligation
					Transition t = selectRandomTransition(state);
					Formula f = obligation.getFormula();
					RemoveVisitor r = new RemoveVisitor(t);
					state.setObligation(f.accept(r).accept(simplifyVisitor));
				}
			} catch (NoSuchElementException e) {
				i--;
			}
		}

		return current;
	}

	/* Random Selection Methods */

	protected Action selectRandomAction(TransitionSystem transitionSystem) {
		return RandomUtil.randomElement(randomness, transitionSystem.actions.getElements());
	}

	private State selectRandomState(TransitionSystem transitionSystem) {
		return RandomUtil.randomElement(randomness, transitionSystem.states.getElements());
	}

	private Transition selectRandomTransition(TransitionSystem transitionSystem) {
		State state = selectRandomState(transitionSystem);
		return selectRandomTransition(state);
	}

	private Transition selectRandomTransition(State state) {
		if (state.getTransitions().isEmpty()) {
			throw new NoSuchElementException();
		}

		return RandomUtil.randomElement(randomness, state.getTransitions());
	}

}
