package edu.tum.motras.legacy.pmtsrc.util.formula;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import com.google.common.collect.Lists;

import edu.tum.motras.legacy.pmtsrc.util.LegacyConfiguration;
import edu.tum.motras.legacy.pmtsrc.util.formula.TreeBoolean.Visitor;

public class QuantifiedBoolean {

	public enum ConversionMethod {
		NAIVE, TSEITIN
	}

	private final LinkedList<Quantifier> quantifiers;
	private final List<List<Literal>> clauses;

	public QuantifiedBoolean(TreeBoolean treeFormula) {
		this(treeFormula, LegacyConfiguration.config.CNF_CONVERSION_METHOD);
	}

	public QuantifiedBoolean(TreeBoolean treeFormula, ConversionMethod method) {
		quantifiers = new LinkedList<>();

		if (treeFormula instanceof Literal) {

			if (((Literal) treeFormula).isTrue()) {
				clauses = new ArrayList<>();
				clauses.add(new ArrayList<Literal>());
				return;
			} else if (((Literal) treeFormula).isFalse()) {
				clauses = new ArrayList<>();
				return;
			}
		}

		switch (method) {
		case TSEITIN:
			TseitinVisitor tVisitor = new TseitinVisitor();
			clauses = treeFormula.accept(tVisitor);
			exists(tVisitor.getRange());
			break;

		case NAIVE:
		default:
			clauses = treeFormula.accept(nVisitor);
			break;
		}

	}

	public void exists(LiteralRange v) {
		if (!v.isEmpty()) {
			quantifiers.addFirst(new Quantifier(Quantifier.EXISTS, v));
		}
	}

	public void forall(LiteralRange v) {
		if (!v.isEmpty()) {
			quantifiers.addFirst(new Quantifier(Quantifier.FORALL, v));
		}
	}

	public List<List<Literal>> getClauses() {
		return clauses;
	}

	public List<Quantifier> getQuantifiers() {
		return Collections.unmodifiableList(quantifiers);
	}

	@Override
	public String toString() {
		return "QuantifiedBoolean [quantifiers=" + quantifiers + ", clauses=" + clauses + "]";
	}

	protected static abstract class ConverterVisitor implements Visitor<List<List<Literal>>> {

		@Override
		public List<List<Literal>> visit(Literal c) {
			List<List<Literal>> clauses = Lists.newArrayList();
			clauses.add(Lists.newArrayList(c));
			return clauses;
		}

		@Override
		public List<List<Literal>> visit(Composition c) {
			if (c.type == Composition.CONJUNCTION) {
				return visitConjunction(c.children);
			}

			return visitDisjunction(c.children);
		}

		private List<List<Literal>> visitConjunction(TreeBoolean c[]) {
			List<List<Literal>> clauses = Lists.newArrayListWithExpectedSize(c.length);

			for (TreeBoolean child : c) {
				clauses.addAll(child.accept(this));
			}

			return clauses;
		}

		protected abstract List<List<Literal>> visitDisjunction(TreeBoolean c[]);
	}

	private final static NaiveVisitor nVisitor = new NaiveVisitor();

	protected static class NaiveVisitor extends ConverterVisitor {
		@Override
		protected List<List<Literal>> visitDisjunction(TreeBoolean c[]) {
			List<List<Literal>> clauses = Lists.newArrayListWithExpectedSize(c.length);
			List<List<Literal>> leftClauses = Lists.newArrayListWithExpectedSize(c.length);
			List<List<Literal>> rightClauses = Lists.newArrayListWithExpectedSize(c.length);

			TreeBoolean tmpLeft, tmpRight;
			TreeBoolean[] tmp = new TreeBoolean[c.length - 1];
			System.arraycopy(c, 1, tmp, 0, tmp.length);

			tmpLeft = c[0];
			tmpRight = TreeBoolean.or(tmp);

			rightClauses = tmpRight.accept(this);
			leftClauses = tmpLeft.accept(this);

			for (List<Literal> leftClause : leftClauses) {
				for (List<Literal> rightClause : rightClauses) {
					List<Literal> clause = Lists.newArrayList();
					clause.addAll(leftClause);
					clause.addAll(rightClause);
					clauses.add(clause);
				}
			}

			return clauses;
		}
	}

	protected static class TseitinVisitor extends ConverterVisitor {
		private final AuxiliaryLiteralFactory manager;

		protected TseitinVisitor() {
			manager = new AuxiliaryLiteralFactory();
			manager.startInitialization();
		}

		public LiteralRange getRange() {
			manager.finishInitialization();
			return manager.getLiteralRange();
		}

		@Override
		protected List<List<Literal>> visitDisjunction(TreeBoolean c[]) {
			List<List<Literal>> clauses = Lists.newArrayListWithExpectedSize(c.length + 1);
			List<Literal> auxClause = Lists.newArrayListWithExpectedSize(c.length);

			for (TreeBoolean child : c) {
				if (child instanceof Literal) {
					auxClause.add((Literal) child);
				} else {
					List<List<Literal>> childClauses = child.accept(this);
					int auxVariable = manager.getVariable();
					Literal auxLiteral = new Literal(auxVariable, true);

					for (List<Literal> clause : childClauses) {
						clause.add(auxLiteral);
						clauses.add(clause);
					}

					auxClause.add(new Literal(auxVariable));
				}
			}

			clauses.add(auxClause);
			return clauses;
		}
	}
}