package edu.tum.xmts;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.annotation.NonNull;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;

import edu.tum.xmts.Obligation.All;
import edu.tum.xmts.Obligation.DisjunctiveMust;
import edu.tum.xmts.Obligation.Must;
import edu.tum.xmts.formula.Formula;

@NonNullByDefault
public class State extends GenericRecord<TransitionSystem> {

	public static class Position {
		public final int X;
		public final int Y;

		Position(int X, int Y) {
			this.X = X;
			this.Y = Y;
		}

		@Override
		public String toString() {
			return "(" + X + "," + Y + ")";
		}
	}

	/* XMTS Properties */

	public static final String PROP_OBLIGATION = "obligation";
	public static final String PROP_POSITION = "position";
	public static final String PROP_TRANSITIONS = "transitions";

	private Obligation obligation;
	private @Nullable Position position;

	private final Table<Action, State, Transition> transitions;
	private final List<Transition> transitionList;

	/* The collections library is missing the necessary null-annotations */
	@SuppressWarnings("null")
	protected State(TransitionSystem context, String id) {
		super(context, id);

		/* Default to true obligation */

		transitions = HashBasedTable.create();
		transitionList = new ArrayList<>();
		obligation = ObligationFactory.create(this, Formula.TRUE);
		position = null;
	}

	public void clearPosition() {
		position = null;
	}

	public Transition createTransition(Action action, State state) {
		checkNotNull(action);
		checkNotNull(state);
		checkArgument(action.context == context);
		checkArgument(state.context == context);
		checkState(!transitions.contains(action, state));

		Transition transition = new Transition(this, action, state);
		transitions.put(action, state, transition);
		transitionList.add(transition);
		propertyChangeSupport.firePropertyChange(PROP_TRANSITIONS, null, transition);
		return transition;
	}

	public Obligation getObligation() {
		return obligation;
	}

	public void setAllObligation() {
		setObligation(ObligationFactory.create(this));
	}

	public void setMustObligation(Collection<Transition> transitions) {
		setObligation(new Obligation.Must(this, transitions));
	}

	public void setDisjunctiveMustObligation(Set<List<Transition>> transitions) {
		setObligation(new Obligation.DisjunctiveMust(this, transitions));
	}

	public void setDisjunctiveMustObligation(Collection<Collection<Transition>> transitions) {
		setObligation(new Obligation.DisjunctiveMust(this, transitions));
	}

	public void setObligation(Set<Set<Transition>> tran) {
		setObligation(ObligationFactory.create(this, tran));
	}

	public void setObligation(Formula formula) {
		setObligation(ObligationFactory.create(this, formula));
	}

	private void setObligation(Obligation obligation) {
		checkArgument(obligation.context == this);

		switch (context.getType()) {
		case MTS:
			checkArgument(obligation instanceof All || obligation instanceof Must);
			break;

		case DMTS:
			checkArgument(obligation instanceof All || obligation instanceof Must
					|| obligation instanceof DisjunctiveMust);
			break;

		case BMTS:
		case PMTS:
		case MTSD:
		case MTSD_DPS:
			break;
		}

		Obligation oldObligation = this.obligation;
		this.obligation = obligation;
		propertyChangeSupport.firePropertyChange(PROP_OBLIGATION, oldObligation, obligation);
	}

	public @Nullable Position getPosition() {
		return position;
	}

	public @Nullable Transition getTransition(Action action, State state) {
		checkNotNull(action);
		checkNotNull(state);
		checkArgument(action.context == context);
		checkArgument(state.context == context);

		return transitions.get(action, state);
	}

	public List<Transition> getTransitions() {
		return transitionList;
	}

	public void removeTransition(Transition transition) {
		checkNotNull(transition);
		checkArgument(transition.context == this);

		transitions.remove(transition.getAction(), transition.getState());
		transitionList.remove(transition);
		obligation.remove(transition);
		propertyChangeSupport.firePropertyChange(PROP_TRANSITIONS, transition, null);
	}

	public void setPosition(int X, int Y) {
		Position oldPosition = position;
		position = new Position(X, Y);
		propertyChangeSupport.firePropertyChange(PROP_POSITION, oldPosition, position);
	}

	@Override
	public boolean isSupportedProperty(@NonNull String propName) {
		if (super.isSupportedProperty(propName)) {
			return true;
		}

		return propName.equals(PROP_OBLIGATION) || propName.equals(PROP_POSITION) || propName.equals(PROP_TRANSITIONS);
	}

	/* The collections library is missing the necessary null-annotations */
	@SuppressWarnings("null")
	public @NonNull Collection<Transition> getTransitions(Action action) {
		checkNotNull(action);
		checkArgument(action.context == context);

		return Collections.unmodifiableCollection(transitions.row(action).values());
	}

	/* The collections library is missing the necessary null-annotations */
	@SuppressWarnings("null")
	public @NonNull Collection<Transition> getTransitions(String action) {
		if (!context.actions.contains(action)) {
			return Collections.emptySet();
		}

		Action actionObj = context.actions.lookup(action);
		return getTransitions(actionObj);
	}

	/* The collections library is missing the necessary null-annotations */
	@SuppressWarnings("null")
	public @NonNull Set<Action> getActions() {
		return Collections.unmodifiableSet(transitions.rowKeySet());
	}
}
