package edu.tum.xmts.formula;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;

import edu.tum.xmts.formula.Formula.Atom;
import edu.tum.xmts.formula.Formula.Conjunction;
import edu.tum.xmts.formula.Formula.Disjunction;
import edu.tum.xmts.formula.Formula.Equivalence;
import edu.tum.xmts.formula.Formula.Implication;
import edu.tum.xmts.formula.Formula.Negation;

public class SimplifyVisitor implements Visitor<Formula> {

	private Set<Formula> visitChildren(Collection<Formula> children) {
		Set<Formula> result = new LinkedHashSet<>();

		for (Formula formula : children) {
			result.add(formula.accept(this));
		}

		return result;
	}

	@Override
	public Formula visit(Atom<?> atom) {
		return atom;
	}

	/* The collections library is missing the necessary null-annotations */
	@SuppressWarnings("null")
	private Collection<Formula> collectConjuncts(Formula formula) {
		if (formula instanceof Conjunction) {
			Collection<Formula> conjuncts = new LinkedHashSet<>();

			for (Formula conjunct : ((Conjunction) formula).children) {
				conjuncts.addAll(collectConjuncts(conjunct));
			}

			return conjuncts;
		}

		return Collections.singleton(formula);
	}

	@Override
	public Formula visit(Conjunction conjunction) {
		Set<Formula> children = visitChildren(collectConjuncts(conjunction));

		children.remove(Formula.TRUE);

		if (children.contains(Formula.FALSE)) {
			return Formula.FALSE;
		}

		switch (children.size()) {
		case 0:
			return Formula.TRUE;

		case 1:
			return children.iterator().next();

		default:
			return Formula.and(children);
		}
	}

	@Override
	public Formula visit(Disjunction disjunction) {
		Set<Formula> children = visitChildren(disjunction.getChildren());

		children.remove(Formula.FALSE);

		if (children.contains(Formula.TRUE)) {
			return Formula.TRUE;
		}

		switch (children.size()) {
		case 0:
			return Formula.FALSE;

		case 1:
			return children.iterator().next();

		default:
			return Formula.or(children);
		}
	}

	@Override
	public Formula visit(Equivalence equivalence) {
		Set<Formula> children = visitChildren(equivalence.getChildren());

		if (children.size() == 1) {
			return Formula.TRUE;
		}

		if (children.remove(Formula.TRUE)) {
			return children.iterator().next();
		}

		if (children.remove(Formula.FALSE)) {
			return children.iterator().next().not();
		}

		Iterator<Formula> iter = children.iterator();

		return new Equivalence(iter.next(), iter.next());
	}

	@Override
	public Formula visit(Implication implication) {
		Formula left = implication.getChildren().get(0).accept(this);
		Formula right = implication.getChildren().get(1).accept(this);

		if (left == Formula.FALSE) {
			return Formula.TRUE;
		}

		if (left == Formula.TRUE) {
			return right;
		}

		if (right == Formula.TRUE) {
			return Formula.TRUE;
		}

		if (right == Formula.FALSE) {
			return left.not().accept(this);
		}

		return left.implies(right);
	}

	@Override
	public Formula visit(Negation negation) {
		Formula child = negation.getChildren().get(0).accept(this);

		if (child == Formula.TRUE) {
			return Formula.FALSE;
		}

		if (child == Formula.FALSE) {
			return Formula.TRUE;
		}

		if (child instanceof Negation) {
			return ((Negation) child).getChildren().get(0);
		}

		return child.not();
	}

}
