package edu.tum.xmts.utils;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.AbstractMap;
import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.ANTLRReaderStream;
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.eclipse.jdt.annotation.NonNull;
import org.eclipse.jdt.annotation.Nullable;

import edu.tum.xmts.Action;
import edu.tum.xmts.Parameter;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.utils.gen.xmtsLexer;
import edu.tum.xmts.utils.gen.xmtsParser;

public class XMTSReader implements TransitionSystemReader {

	public static class MalformedInputException extends Exception {

		public MalformedInputException(@Nullable String string) {
			super(string);
		}

		public MalformedInputException() {
			super();
		}

		private static final long serialVersionUID = -349103923181242025L;

	}

	private static Logger logger = Logger.getLogger(XMTSReader.class.getName());

	private static <T> T handleIllegalCommonTree(CommonTree tree) throws MalformedInputException {
		logger.warning("Illegal xmts token: " + tree.getType() + " " + tree.getText());
		throw new MalformedInputException("Illegal xmts token: " + tree.getType() + " " + tree.getText());
	}

	private Deque<Map.Entry<Object, List<?>>> initQueue;
	private @Nullable String initialState;
	private TransitionSystem transitionSystem;

	/* Input */

	public XMTSReader() {
		initQueue = new LinkedList<>();
		transitionSystem = new TransitionSystem("DUMMY");
	}

	private void initialise(Action action, CommonTree property) throws MalformedInputException {
		switch (property.getType()) {
		case xmtsParser.LABEL:
			action.setLabel(property.getText().substring(1, property.getText().length() - 1));
			break;

		case xmtsParser.REQUIREMENT:
			@Nullable
			CommonTree onlyChild = (CommonTree) property.getChild(0);

			if (onlyChild == null) {
				throw new MalformedInputException("Requirement for action " + action.getId()
						+ " defined, but no formula found.");
			}

			action.setRequirement(parseFormula(onlyChild, null));
			break;

		case xmtsParser.COST:
			action.setRunningCost(new Integer(property.getText()));
			break;

		default:
			handleIllegalCommonTree(property);
		}
	}

	private static void initialise(Parameter parameter, CommonTree property) throws MalformedInputException {
		switch (property.getType()) {
		case xmtsParser.LABEL:
			parameter.setLabel(property.getText().substring(1, property.getText().length() - 1));
			break;

		case xmtsParser.INVESTMENT:
			parameter.setInvestmentCost(new Integer(property.getText()));
			break;

		default:
			handleIllegalCommonTree(property);
		}
	}

	private void initialise(State state, CommonTree property) throws MalformedInputException {
		switch (property.getType()) {
		case xmtsParser.LABEL:
			state.setLabel(property.getText().substring(1, property.getText().length() - 1));
			break;

		/* Delay obligation construction until all transitions are constructed */
		case xmtsParser.OBLIGATION:
			Object child = property.getChild(0);
			if (child != null) {
				postponeInitialisation(state, child);
			}
			break;

		case xmtsParser.ALL:
			state.setAllObligation();
			break;

		case xmtsParser.FORMULA:
			state.setObligation(parseFormula(property, state));
			break;

		case xmtsParser.POSITION:
			int X = Integer.parseInt(((CommonTree) property.getChild(0)).getText());
			int Y = Integer.parseInt(((CommonTree) property.getChild(1)).getText());
			state.setPosition(X, Y);
			break;

		case xmtsParser.TRANSITION:
			Action action = transitionSystem.actions.lookup(((CommonTree) property.getChild(0)).getText());
			State target = transitionSystem.states.lookup(((CommonTree) property.getChild(1)).getText());
			Transition transition = state.createTransition(action, target);

			CommonTree childProperty = (CommonTree) property.getChild(2);
			if (childProperty != null) {
				initialise(transition, childProperty);
			}

			break;

		default:
			handleIllegalCommonTree(property);
		}
	}

	private <T> void initialise(T object, @Nullable List<?> properties) throws MalformedInputException {
		if (properties == null) {
			return;
		}

		for (Object property : properties) {
			if (property == null) {
				logger.info("Skipping null property for " + object);
				continue;
			}

			if (object instanceof TransitionSystem) {
				initialise((TransitionSystem) object, (CommonTree) property);
				continue;
			}

			if (object instanceof Parameter) {
				initialise((Parameter) object, (CommonTree) property);
				continue;
			}

			if (object instanceof Action) {
				initialise((Action) object, (CommonTree) property);
				continue;
			}

			if (object instanceof State) {
				initialise((State) object, (CommonTree) property);
				continue;
			}

			if (object instanceof Transition) {
				initialise((Transition) object, (CommonTree) property);
				continue;
			}

			logger.info("Unknown object in initialise " + object);
		}
	}

	private static void initialise(Transition transition, CommonTree property) throws MalformedInputException {
		switch (property.getType()) {
		case xmtsParser.LABEL:
			transition.setLabel(property.getText().substring(1, property.getText().length() - 1));
			break;

		case xmtsParser.INT_CON:
			int lower = Integer.parseInt(((CommonTree) property.getChild(0)).getText());
			int upper = Integer.parseInt(((CommonTree) property.getChild(1)).getText());
			transition.setDuration(Transition.Duration.Type.CONTROLLABLE, lower, upper);
			break;

		case xmtsParser.INT_UNCON:
			int lower_ = Integer.parseInt(((CommonTree) property.getChild(0)).getText());
			int upper_ = Integer.parseInt(((CommonTree) property.getChild(1)).getText());
			transition.setDuration(Transition.Duration.Type.UNCONTROLLABLE, lower_, upper_);
			break;

		default:
			handleIllegalCommonTree(property);
		}
	}

	private void initialise(TransitionSystem transitionSystem, @Nullable CommonTree property)
			throws MalformedInputException {
		if (property == null) {
			return;
		}

		String propertyText = property.getText();

		if (propertyText == null) {
			throw new MalformedInputException("Missing text");
		}

		switch (property.getType()) {
		case xmtsParser.LABEL:
			transitionSystem.setLabel(propertyText.substring(1, propertyText.length() - 1));
			break;

		case xmtsParser.INITIAL:
			initialState = propertyText;
			break;

		case xmtsParser.PARAMETER:
			Parameter parameter = transitionSystem.parameters.create(propertyText);
			initialise(parameter, property.getChildren());
			break;

		case xmtsParser.ACTION:
			Action action = transitionSystem.actions.create(propertyText);
			initialise(action, property.getChildren());
			break;

		case xmtsParser.STATE:
			State state = transitionSystem.states.create(propertyText);
			List<?> children = property.getChildren();
			if (children != null) {
				postponeInitialisation(state, children);
			}
			break;

		default:
			handleIllegalCommonTree(property);
		}
	}

	@SuppressWarnings("unchecked")
	private Formula parseFormula(CommonTree formula, @Nullable State state) throws MalformedInputException {
		if (formula.getType() != xmtsParser.FORMULA) {
			handleIllegalCommonTree(formula);
		}

		if (formula.getChildCount() == 1) {
			@Nullable
			CommonTree onlyChild = (CommonTree) formula.getChild(0);

			if (onlyChild == null) {
				throw new MalformedInputException("One child claimed, null returned");
			}

			return parseUnit(onlyChild, state);
		}

		List<CommonTree> children = new LinkedList<>();
		children.addAll((Collection<? extends CommonTree>) formula.getChildren());

		switch (children.remove(1).getType()) {
		case xmtsParser.AND:
			return Formula.and(parseUnit(children, state));

		case xmtsParser.OR:
			return Formula.or(parseUnit(children, state));

		case xmtsParser.IMP:
			return new Formula.Implication(parseUnit(children.get(0), state), parseUnit(children.get(1), state));

		case xmtsParser.EQV:
			return new Formula.Equivalence(parseUnit(children.get(0), state), parseUnit(children.get(1), state));

		default:
			return handleIllegalCommonTree(formula);
		}
	}

	private Collection<Formula> parseUnit(Collection<CommonTree> units, @Nullable State state)
			throws MalformedInputException {
		Collection<Formula> formulas = new LinkedList<>();

		for (CommonTree unit : units) {
			formulas.add(parseUnit(unit, state));
		}

		return formulas;
	}

	private Formula parseUnit(CommonTree unit, @Nullable State state) throws MalformedInputException {
		switch (unit.getType()) {
		case xmtsParser.UNIT:
			CommonTree child = (CommonTree) unit.getChild(0);
			if (child != null) {
				return parseFormula(child, state);
			}
			throw new MalformedInputException("Unit is empty");

		case xmtsParser.ATOM_C:
			return Boolean.parseBoolean(unit.getText()) ? Formula.TRUE : Formula.FALSE;

		case xmtsParser.ATOM_P:
			return new Formula.Atom<>(transitionSystem.parameters.lookup(unit.getText()));

		case xmtsParser.ATOM:
			if (state == null) {
				throw new MalformedInputException("There is no state referenced");
			}

			Transition transition = state.getTransition(transitionSystem.actions.lookup(unit.getChild(0).getText()),
					transitionSystem.states.lookup(unit.getChild(1).getText()));

			if (transition == null) {
				throw new MalformedInputException("Non-existing transition referenced");
			}

			return new Formula.Atom<>(transition);

		case xmtsParser.NOT:
			CommonTree child2 = (CommonTree) unit.getChild(0);
			if (child2 != null) {
				return parseUnit(child2, state).not();
			}
			throw new MalformedInputException("Negation is empty");

		default:
			return handleIllegalCommonTree(unit);
		}
	}

	private void postponeInitialisation(Object object, List<?> properties) {
		initQueue.addLast(new AbstractMap.SimpleEntry<Object, List<?>>(object, properties));
	}

	@SuppressWarnings("null")
	private void postponeInitialisation(Object object, Object property) {
		initQueue.addLast(new AbstractMap.SimpleEntry<Object, List<?>>(object, Collections.singletonList(property)));
	}

	private void resumeInitialisation() throws MalformedInputException {
		while (!initQueue.isEmpty()) {
			Map.Entry<Object, List<?>> pair = initQueue.removeFirst();
			initialise(pair.getKey(), pair.getValue());
		}
	}

	private TransitionSystem walkInput(CommonTree xmts) throws MalformedInputException {
		initialState = null;
		initQueue = new LinkedList<>();
		String text = xmts.getText();

		if (text == null) {
			throw new MalformedInputException();
		}

		transitionSystem = new TransitionSystem(text);

		switch (xmts.getType()) {
		case xmtsParser.MTS:
			transitionSystem.setType(Type.MTS);
			break;

		case xmtsParser.DMTS:
			transitionSystem.setType(Type.DMTS);
			break;

		case xmtsParser.BMTS:
			transitionSystem.setType(Type.DMTS);
			break;

		case xmtsParser.PMTS:
			transitionSystem.setType(Type.PMTS);
			break;

		case xmtsParser.MTSD:
			transitionSystem.setType(Type.MTSD);
			break;

		case xmtsParser.MTSD_DPS:
			transitionSystem.setType(Type.MTSD_DPS);
			break;

		default:
			handleIllegalCommonTree(xmts);
		}

		initialise(transitionSystem, xmts.getChildren());
		resumeInitialisation();

		if (initialState != null) {
			transitionSystem.setInitialState(transitionSystem.states.lookup(initialState));
		}

		TransitionSystem result = transitionSystem;

		transitionSystem = new TransitionSystem("DUMMY");
		initialState = null;
		initQueue.clear();

		return result;
	}

	private static xmtsParser createParser(CharStream input) {
		xmtsLexer lex;
		CommonTokenStream tokens;

		lex = new xmtsLexer(input);
		tokens = new CommonTokenStream(lex);
		return new xmtsParser(tokens);
	}

	private TransitionSystem read(CharStream input) throws MalformedInputException {
		try {
			xmtsParser.xmts_return xmts = createParser(input).xmts();

			CommonTree xmts2 = (CommonTree) xmts.getTree();

			if (xmts2 == null) {
				throw new MalformedInputException("tree is null");
			}

			return walkInput(xmts2);
		} catch (RecognitionException e) {
			throw new MalformedInputException(e.getMessage());
		}
	}

	@Override
	public @NonNull TransitionSystem read(File f) throws IOException, MalformedInputException {
		return read(new ANTLRFileStream(f.getAbsolutePath()));
	}

	@Override
	public TransitionSystem read(InputStream i) throws IOException, MalformedInputException {
		return read(new ANTLRInputStream(i));
	}

	@Override
	public TransitionSystem read(Reader r) throws IOException, MalformedInputException {
		return read(new ANTLRReaderStream(r));
	}

	@Override
	public TransitionSystem read(String s) throws MalformedInputException {
		return read(new ANTLRStringStream(s));
	}

	private List<TransitionSystem> readList(CharStream input) throws MalformedInputException {
		List<TransitionSystem> systems = new LinkedList<>();

		try {
			xmtsParser.xmts_list_return xmts_list = createParser(input).xmts_list();

			for (Object xmts : ((CommonTree) xmts_list.getTree()).getChildren()) {
				if (xmts == null) {
					continue;
				}

				systems.add(walkInput(((CommonTree) xmts)));
			}

			return systems;
		} catch (RecognitionException e) {
			throw new MalformedInputException(e.getMessage());
		}
	}

	public List<TransitionSystem> readList(String s) throws MalformedInputException {
		return readList(new ANTLRStringStream(s));
	}

	public List<TransitionSystem> readList(InputStream i) throws IOException, MalformedInputException {
		return readList(new ANTLRInputStream(i));
	}

	public List<TransitionSystem> readList(Reader r) throws IOException, MalformedInputException {
		return readList(new ANTLRReaderStream(r));
	}

	private Formula readFormula(TransitionSystem system, @Nullable State context, CharStream input)
			throws MalformedInputException {
		try {
			transitionSystem = system;
			xmtsParser.formula_return formula = createParser(input).formula();
			CommonTree formulaTree = (CommonTree) formula.getTree();
			if (formulaTree == null) {
				throw new MalformedInputException("FormulaTree is null");
			}
			Formula f = parseFormula(formulaTree, context);
			transitionSystem = new TransitionSystem("DUMMY");
			return f;
		} catch (RecognitionException e) {
			throw new MalformedInputException(e.getMessage());
		}
	}

	public Formula readFormula(TransitionSystem context, String s) throws MalformedInputException {
		// Work-around for ANTLR: "(" + s.trim() + ")"
		return readFormula(context, null, new ANTLRStringStream("(" + s.trim() + ")"));
	}

	public Formula readFormula(State context, String s) throws MalformedInputException {
		// Work-around for ANTLR: "(" + s.trim() + ")"
		return readFormula(context.getContext(), context, new ANTLRStringStream("(" + s.trim() + ")"));
	}

}
