package edu.tum.motras.algorithms.junit;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.Collection;
import java.util.Collections;
import java.util.Deque;

import org.junit.Before;
import org.junit.Test;

import com.google.common.collect.ImmutableSet;

import edu.tum.motras.algorithms.Algorithm;
import edu.tum.motras.algorithms.CompositionMTS;
import edu.tum.motras.algorithms.MaxImplementation;
import edu.tum.motras.algorithms.ModalRefinementFixedPoint;
import edu.tum.motras.algorithms.util.Util;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.utils.XMTSReader;
import edu.tum.xmts.utils.XMTSReader.MalformedInputException;

public class CompositionMTSTest {

	CompositionMTS compositionMTS;

	@Before
	public void setUp() throws Exception {
		compositionMTS = new CompositionMTS("a");
	}

	@Test
	public void testApply() throws MalformedInputException {
		String xmts1 = "mts test1[action a action b " + "state s1 [obligation a.s3 b->s2 a->s3]" + "state s2 "
				+ "state s3 [obligation true b->s3]]";

		String xmts2 = "mts test2[action a action c " + "state t1 [obligation all  c->t2 a->t3]" + "state t2 "
				+ "state t3 [obligation true c->t3]]";

		XMTSReader reader = new XMTSReader();
		TransitionSystem system1 = reader.read(xmts1);
		TransitionSystem system2 = reader.read(xmts2);

		Collection<Transition> aTran;
		Collection<Transition> bTran;
		Collection<Transition> cTran;
		Collection<Transition> mustTran;

		State s1_t1 = compositionMTS.apply(system1.states.lookup("s1"), system2.states.lookup("t1"));

		/* Test TransitionSystem */
		TransitionSystem compositionSystem = s1_t1.getContext();
		assertEquals(Type.MTS, compositionSystem.getType());
		assertEquals(compositionSystem.parameters.getElements().toString(), 0, compositionSystem.parameters
				.getElements().size());
		assertEquals(compositionSystem.actions.getElements().toString(), 3, compositionSystem.actions.getElements()
				.size());
		assertEquals(compositionSystem.states.getElements().toString(), 5, compositionSystem.states.getElements()
				.size());

		/* Test State {s1,t1} */
		aTran = s1_t1.getTransitions("a");
		bTran = s1_t1.getTransitions("b");
		cTran = s1_t1.getTransitions("c");

		assertEquals(aTran.toString(), 1, aTran.size());
		assertEquals(bTran.toString(), 1, bTran.size());
		assertEquals(cTran.toString(), 1, cTran.size());

		mustTran = Util.getMustTransitions(s1_t1);

		State s3_t3 = aTran.iterator().next().getState();
		State s2_t1 = bTran.iterator().next().getState();
		State s1_t2 = cTran.iterator().next().getState();

		assertEquals(ImmutableSet.of(aTran.iterator().next(), cTran.iterator().next()), mustTran);

		/* Test State {s3, t3} */
		aTran = s3_t3.getTransitions("a");
		bTran = s3_t3.getTransitions("b");
		cTran = s3_t3.getTransitions("c");

		assertEquals(aTran.toString(), 0, aTran.size());
		assertEquals(bTran.toString(), 1, bTran.size());
		assertEquals(cTran.toString(), 1, cTran.size());

		assertEquals(Collections.emptySet(), Util.getMustTransitions(s3_t3));
		assertEquals(s3_t3, bTran.iterator().next().getState());
		assertEquals(s3_t3, cTran.iterator().next().getState());

		/* Test State {s2, t1} */
		aTran = s2_t1.getTransitions("a");
		bTran = s2_t1.getTransitions("b");
		cTran = s2_t1.getTransitions("c");

		assertEquals(aTran.toString(), 0, aTran.size());
		assertEquals(bTran.toString(), 0, bTran.size());
		assertEquals(cTran.toString(), 1, cTran.size());

		mustTran = Util.getMustTransitions(s2_t1);

		State s2_t2 = cTran.iterator().next().getState();

		assertEquals(Collections.singleton(cTran.iterator().next()), Util.getMustTransitions(s2_t1));

		/* Test State {s1, t2} */
		aTran = s1_t2.getTransitions("a");
		bTran = s1_t2.getTransitions("b");
		cTran = s1_t2.getTransitions("c");

		assertEquals(aTran.toString(), 0, aTran.size());
		assertEquals(bTran.toString(), 1, bTran.size());
		assertEquals(cTran.toString(), 0, cTran.size());

		assertEquals(Collections.emptySet(), Util.getMustTransitions(s1_t2));
		assertEquals(s2_t2, bTran.iterator().next().getState());

		/* Test State {s2, t2} */
		aTran = s2_t2.getTransitions("a");
		bTran = s2_t2.getTransitions("b");
		cTran = s2_t2.getTransitions("c");

		assertEquals(aTran.toString(), 0, aTran.size());
		assertEquals(bTran.toString(), 0, bTran.size());
		assertEquals(cTran.toString(), 0, cTran.size());
	}

	@Test
	public void testApplyRandom() throws Exception {
		Algorithm<Boolean> modalRefinement = new ModalRefinementFixedPoint();
		CompositionMTS composition1 = new CompositionMTS("a, a0, a0");
		CompositionMTS composition2 = new CompositionMTS("a, a0, a0");
		Deque<TransitionSystem> transitionSystems = AlgorithmTestUtil.generate(Type.MTS);
		MaxImplementation maxImplementation1 = new MaxImplementation();
		MaxImplementation maxImplementation2 = new MaxImplementation();

		assertEquals(ImmutableSet.of("a0", "a"), composition1.getSynchronizingAlphabet());

		while (transitionSystems.size() > 1) {
			TransitionSystem transitionSystem1 = transitionSystems.pop();
			TransitionSystem transitionSystem2 = transitionSystems.pop();

			State abstract1 = transitionSystem1.states.getElements().get(0);
			State abstract2 = transitionSystem2.states.getElements().get(0);

			State implementation1 = maxImplementation1.apply(abstract1);
			State implementation2 = maxImplementation2.apply(abstract2);

			State compAbstract = composition1.apply(abstract1, abstract2);
			State compImplementation = composition2.apply(implementation1, implementation2);

			assertTrue(modalRefinement.apply(implementation1, abstract1));
			assertTrue(modalRefinement.apply(implementation2, abstract2));
			assertTrue(modalRefinement.apply(compImplementation, compAbstract));
		}
	}

	@Test
	public void testGetArity() {
		AlgorithmTestUtil.testGetArity(compositionMTS, 2);
	}

	@Test
	public void testGetSupportedTypes() {
		AlgorithmTestUtil.testGetSupportedTypes(compositionMTS, ImmutableSet.of(Type.MTS));
	}
}
