package edu.tum.motras.algorithms.junit;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import com.google.common.collect.ImmutableSet;

import edu.tum.motras.algorithms.DeterministicHullBMTS;
import edu.tum.motras.algorithms.ModalRefinementQSat;
import edu.tum.xmts.Action;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.utils.XMTSReader;
import edu.tum.xmts.utils.XMTSReader.MalformedInputException;

public class DeterministicHullBMTSTest {

	private DeterministicHullBMTS deterministicHullBMTS;

	@Before
	public void setUp() throws Exception {
		deterministicHullBMTS = new DeterministicHullBMTS();
	}

	@Test
	public void testApply() throws MalformedInputException {
		String xmts = "mts test[action a action b " + "state s1 [obligation a.s2 a->s2 a->s3 b->s4]"
				+ "state s2 [obligation all  a->s3 a->s5]" + "state s3 [obligation b.s6 b->s4 b->s6]"
				+ "state s4 [obligation all  a->s4]" + "state s5 [obligation all  b->s6]"
				+ "state s6 [obligation true a->s4]]";

		XMTSReader reader = new XMTSReader();
		TransitionSystem system = reader.read(xmts);

		Collection<Transition> aTran;
		Collection<Transition> bTran;

		State s1 = deterministicHullBMTS.apply(new State[] { system.states.lookup("s1") });

		/* Test TransitionSystem */
		TransitionSystem deterministicSystem = s1.getContext();
		assertEquals(Type.BMTS, deterministicSystem.getType());
		assertEquals(deterministicSystem.parameters.getElements().toString(), 0, deterministicSystem.parameters
				.getElements().size());
		assertEquals(deterministicSystem.actions.getElements().toString(), 2, deterministicSystem.actions.getElements()
				.size());
		assertEquals(deterministicSystem.states.getElements().toString(), 5, deterministicSystem.states.getElements()
				.size());

		/* Test State {s1} */
		aTran = s1.getTransitions("a");
		bTran = s1.getTransitions("b");

		assertTrue(aTran.size() == 1);
		assertTrue(bTran.size() == 1);

		State s2_s3 = aTran.iterator().next().getState();
		State s4 = bTran.iterator().next().getState();

		/* Test State {s2, s3} */
		aTran = s2_s3.getTransitions("a");
		bTran = s2_s3.getTransitions("b");

		assertTrue(aTran.size() == 1);
		assertTrue(bTran.size() == 1);

		State s3_s5 = aTran.iterator().next().getState();
		State s4_s6 = bTran.iterator().next().getState();

		/* Test State {s3, s5} */
		aTran = s3_s5.getTransitions("a");
		bTran = s3_s5.getTransitions("b");

		assertTrue(aTran.size() == 0);
		assertTrue(bTran.size() == 1);

		assertTrue(s4_s6 == bTran.iterator().next().getState());

		/* Test State {s4, s6} */
		aTran = s4_s6.getTransitions("a");
		bTran = s4_s6.getTransitions("b");

		assertTrue(aTran.size() == 1);
		assertTrue(bTran.size() == 0);

		assertTrue(s4 == aTran.iterator().next().getState());

		/* Test State {s4} */
		aTran = s4.getTransitions("a");
		bTran = s4.getTransitions("b");

		assertTrue(aTran.size() == 1);
		assertTrue(bTran.size() == 0);

		assertTrue(s4 == aTran.iterator().next().getState());
	}

	@Test
	public void testApplyRandom() throws Exception {
		ModalRefinementQSat modalRefinement = new ModalRefinementQSat();
		List<TransitionSystem> transitionSystems = new LinkedList<>();

		transitionSystems.addAll(AlgorithmTestUtil.generate(Type.MTS));
		transitionSystems.addAll(AlgorithmTestUtil.generate(Type.DMTS));
		transitionSystems.addAll(AlgorithmTestUtil.generate(Type.BMTS));

		for (TransitionSystem system : transitionSystems) {
			TransitionSystem deterministicSystem;
			State state, deterministicState;

			state = system.states.getElements().get(0);
			deterministicState = deterministicHullBMTS.apply(state);
			deterministicSystem = deterministicState.getContext();

			assertTrue(modalRefinement.apply(state, deterministicState));

			for (State detState : deterministicSystem.states.getElements()) {
				for (Action action : detState.getActions()) {
					assertTrue(detState.getTransitions(action).size() <= 1);
				}
			}
		}
	}

	@Test
	public void testGetArity() {
		AlgorithmTestUtil.testGetArity(deterministicHullBMTS, 1);
	}

	@Test
	public void testGetSupportedTypes() {
		AlgorithmTestUtil.testGetSupportedTypes(deterministicHullBMTS, ImmutableSet.of(Type.MTS, Type.DMTS, Type.BMTS));
	}
}
