package edu.tum.motras.algorithms.junit;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.Collection;

import org.junit.Before;
import org.junit.Test;

import com.google.common.collect.ImmutableSet;

import edu.tum.motras.algorithms.Algorithm;
import edu.tum.motras.algorithms.DeterministicHullMTS;
import edu.tum.motras.algorithms.ModalRefinementFixedPoint;
import edu.tum.motras.algorithms.util.Util;
import edu.tum.xmts.Action;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.utils.XMTSReader;
import edu.tum.xmts.utils.XMTSReader.MalformedInputException;

public class DeterministicHullMTSTest {

	private DeterministicHullMTS deterministicHullMTS;

	@Before
	public void setUp() throws Exception {
		deterministicHullMTS = new DeterministicHullMTS();
	}

	@Test
	public void testApply() throws MalformedInputException {
		String xmts = "mts test[action a action b " + "state s1 [obligation a.s2 a->s2 a->s3 b->s4]"
				+ "state s2 [obligation all  a->s3 a->s5]" + "state s3 [obligation b.s6 b->s4 b->s6]"
				+ "state s4 [obligation all  a->s4]" + "state s5 [obligation all  b->s6]"
				+ "state s6 [obligation true a->s4]]";

		XMTSReader reader = new XMTSReader();
		TransitionSystem system = reader.read(xmts);

		Collection<Transition> aTran;
		Collection<Transition> bTran;

		State s1 = deterministicHullMTS.apply(new State[] { system.states.lookup("s1") });

		/* Test TransitionSystem */
		TransitionSystem deterministicSystem = s1.getContext();
		assertEquals(Type.MTS, deterministicSystem.getType());
		assertEquals(deterministicSystem.parameters.getElements().toString(), 0, deterministicSystem.parameters
				.getElements().size());
		assertEquals(deterministicSystem.actions.getElements().toString(), 2, deterministicSystem.actions.getElements()
				.size());
		assertEquals(deterministicSystem.states.getElements().toString(), 5, deterministicSystem.states.getElements()
				.size());

		/* Test State {s1} */
		aTran = s1.getTransitions("a");
		bTran = s1.getTransitions("b");

		assertTrue(aTran.size() == 1);
		assertTrue(bTran.size() == 1);
		assertEquals(1, Util.getMustTransitions(s1).size());
		assertTrue(Util.getMustTransitions(s1).contains(aTran.iterator().next()));

		State s2_s3 = aTran.iterator().next().getState();
		State s4 = bTran.iterator().next().getState();

		/* Test State {s2, s3} */
		aTran = s2_s3.getTransitions("a");
		bTran = s2_s3.getTransitions("b");

		assertTrue(aTran.size() == 1);
		assertTrue(bTran.size() == 1);
		assertTrue(Util.getMustTransitions(s2_s3).isEmpty());

		State s3_s5 = aTran.iterator().next().getState();
		State s4_s6 = bTran.iterator().next().getState();

		/* Test State {s3, s5} */
		aTran = s3_s5.getTransitions("a");
		bTran = s3_s5.getTransitions("b");

		assertTrue(aTran.size() == 0);
		assertTrue(bTran.size() == 1);
		assertTrue(Util.getMustTransitions(s3_s5).containsAll(s3_s5.getTransitions()));
		assertTrue(s3_s5.getTransitions().containsAll(Util.getMustTransitions(s3_s5)));

		assertTrue(s4_s6 == bTran.iterator().next().getState());

		/* Test State {s4, s6} */
		aTran = s4_s6.getTransitions("a");
		bTran = s4_s6.getTransitions("b");

		assertTrue(aTran.size() == 1);
		assertTrue(bTran.size() == 0);
		assertTrue(Util.getMustTransitions(s4_s6).isEmpty());

		assertTrue(s4 == aTran.iterator().next().getState());

		/* Test State {s4} */
		aTran = s4.getTransitions("a");
		bTran = s4.getTransitions("b");

		assertTrue(aTran.size() == 1);
		assertTrue(bTran.size() == 0);
		assertTrue(Util.getMustTransitions(s4).containsAll(s4.getTransitions()));
		assertTrue(s4.getTransitions().containsAll(Util.getMustTransitions(s4)));

		assertTrue(s4 == aTran.iterator().next().getState());
	}

	@Test
	public void testApplyRandom() throws Exception {
		Algorithm<Boolean> modalRefinement = new ModalRefinementFixedPoint();

		for (TransitionSystem system : AlgorithmTestUtil.generate(Type.MTS)) {
			TransitionSystem deterministicSystem;
			State state, deterministicState;

			state = system.states.getElements().get(0);
			deterministicState = deterministicHullMTS.apply(state);
			deterministicSystem = deterministicState.getContext();

			assertTrue(modalRefinement.apply(state, deterministicState));

			for (State detState : deterministicSystem.states.getElements()) {
				for (Action action : detState.getActions()) {
					assertTrue(detState.getTransitions(action).size() <= 1);
				}
			}
		}
	}

	@Test
	public void testGetArity() {
		AlgorithmTestUtil.testGetArity(deterministicHullMTS, 1);
	}

	@Test
	public void testGetSupportedTypes() {
		AlgorithmTestUtil.testGetSupportedTypes(deterministicHullMTS, ImmutableSet.of(Type.MTS));
	}
}
