package edu.tum.xmts.junit;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.util.List;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import edu.tum.xmts.Action;
import edu.tum.xmts.Obligation;
import edu.tum.xmts.Parameter;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.formula.Formula.Atom;
import edu.tum.xmts.utils.XMTSReader;
import edu.tum.xmts.utils.XMTSReader.MalformedInputException;

public class XMTSReaderTest {

	XMTSReader reader;

	@Before
	public void setUp() throws Exception {
		reader = new XMTSReader();
	}

	@After
	public void tearDown() {
		reader = null;
	}

	@SuppressWarnings("null")
	@Test
	public void testReadFile() throws IOException, MalformedInputException {
		File examplesDir = new File("../xmts-specification/examples");

		for (File xmtsFile : examplesDir.listFiles((d, s) -> s.endsWith(".xmts"))) {
			reader.read(xmtsFile);
		}
	}

	@SuppressWarnings("null")
	@Test
	public void testTrafficFile() throws IOException, MalformedInputException {
		File examplesDir = new File("../xmts-specification/examples/traffic");

		for (File xmtsFile : examplesDir.listFiles((d, s) -> s.endsWith(".xmts"))) {
			reader.read(xmtsFile);
		}
	}

	@SuppressWarnings("null")
	@Test
	public void testCoffeeFile() throws IOException, MalformedInputException {
		File examplesDir = new File("../xmts-specification/examples/coffee and tea");

		for (File xmtsFile : examplesDir.listFiles((d, s) -> s.endsWith(".xmts"))) {
			reader.read(xmtsFile);
		}
	}

	@Test
	public void testReadFormula() throws MalformedInputException {
		TransitionSystem system = new TransitionSystem("test");
		system.setType(Type.MTSD_DPS);

		Parameter p = system.parameters.create("p");
		Action a = system.actions.create("a");

		State s1 = system.states.create("s1");
		State s2 = system.states.create("s2");
		Transition t = s1.createTransition(a, s2);

		assertEquals(Boolean.TRUE, ((Atom<?>) reader.readFormula(s1, " true")).getPayload());
		assertEquals(t, ((Atom<?>) reader.readFormula(s1, "a.s2")).getPayload());
		assertEquals(t, ((Atom<?>) reader.readFormula(s1, "(a.s2)")).getPayload());

		assertEquals(Boolean.TRUE, ((Atom<?>) reader.readFormula(system, " true")).getPayload());
		assertEquals(p, ((Atom<?>) reader.readFormula(s1, "p")).getPayload());
		assertEquals(p, ((Atom<?>) reader.readFormula(s1, "(p)")).getPayload());
	}

	@Test
	public void testReadString() throws MalformedInputException {
		String X = "pmts X [action a " + "state S  [a -> S1]" + "state S2 " + "state S1 [obligation a.S a -> S]"
				+ "state T  [obligation a.T1 & (a.T2 |  a.T3) a -> T1 a -> T2 a-> T3]"
				+ "state T1 [obligation a.T1 & (a.T2 | !a.T3) a -> T1 a -> T2 a-> T3]"
				+ "state T2 [obligation all a -> T1]" + "state T3 [obligation a.T2 |  a.T3 a -> T1 a -> T2 a-> T3]"
				+ "]";

		TransitionSystem system = reader.read(X);

		assertTrue(system.getId().equals("X"));
		assertTrue(system.getType() == Type.PMTS);

		State S = system.states.lookup("S");
		State S1 = system.states.lookup("S1");
		State S2 = system.states.lookup("S2");
		State T = system.states.lookup("T");
		State T1 = system.states.lookup("T1");
		State T2 = system.states.lookup("T2");
		State T3 = system.states.lookup("T3");

		/* Check Id invariants */
		assertTrue(system.actions.lookup("a").getId().equals("a"));
		assertTrue(S.getId().equals("S"));
		assertTrue(S1.getId().equals("S1"));
		assertTrue(S2.getId().equals("S2"));
		assertTrue(T.getId().equals("T"));
		assertTrue(T1.getId().equals("T1"));
		assertTrue(T2.getId().equals("T2"));
		assertTrue(T3.getId().equals("T3"));

		/* Check Obligation Construction */
		assertTrue(S.getObligation() instanceof Obligation.Must);
		assertTrue(S1.getObligation() instanceof Obligation.Must);
		assertTrue(S2.getObligation() instanceof Obligation.Must);
		assertTrue(T.getObligation() instanceof Obligation.DisjunctiveMust);
		assertTrue(T1.getObligation() instanceof Obligation.Generic);
		assertTrue(T2.getObligation() instanceof Obligation.All);
		assertTrue(T3.getObligation() instanceof Obligation.DisjunctiveMust);

	}

	@Test
	public void testReadListString() throws MalformedInputException {
		String X = "pmts X [action a " + "state S  [a -> S1]" + "state S2 " + "state S1 [obligation a.S a -> S]"
				+ "state T  [obligation a.T1 & (a.T2 |  a.T3) a -> T1 a -> T2 a-> T3]"
				+ "state T1 [obligation a.T1 & (a.T2 | !a.T3) a -> T1 a -> T2 a-> T3]"
				+ "state T2 [obligation all a -> T1]" + "state T3 [obligation a.T2 |  a.T3 a -> T1 a -> T2 a-> T3]"
				+ "]" + " pmts Y [action a " + "state S  [a -> S1]" + "state S2 " + "state S1 [obligation a.S a -> S]"
				+ "state T  [obligation a.T1 & (a.T2 |  a.T3) a -> T1 a -> T2 a-> T3]"
				+ "state T1 [obligation a.T1 & (a.T2 | !a.T3) a -> T1 a -> T2 a-> T3]"
				+ "state T2 [obligation all a -> T1]" + "state T3 [obligation a.T2 |  a.T3 a -> T1 a -> T2 a-> T3]"
				+ "]";

		List<TransitionSystem> systems = reader.readList(X);

		assertTrue(systems.size() == 2);

		TransitionSystem system = systems.get(0);

		assertTrue(system.getId().equals("X"));
		assertTrue(system.getType() == Type.PMTS);

		State S = system.states.lookup("S");
		State S1 = system.states.lookup("S1");
		State S2 = system.states.lookup("S2");
		State T = system.states.lookup("T");
		State T1 = system.states.lookup("T1");
		State T2 = system.states.lookup("T2");
		State T3 = system.states.lookup("T3");

		/* Check Id invariants */
		assertTrue(system.actions.lookup("a").getId().equals("a"));
		assertTrue(S.getId().equals("S"));
		assertTrue(S1.getId().equals("S1"));
		assertTrue(S2.getId().equals("S2"));
		assertTrue(T.getId().equals("T"));
		assertTrue(T1.getId().equals("T1"));
		assertTrue(T2.getId().equals("T2"));
		assertTrue(T3.getId().equals("T3"));

		/* Check Obligation Construction */
		assertTrue(S.getObligation() instanceof Obligation.Must);
		assertTrue(S1.getObligation() instanceof Obligation.Must);
		assertTrue(S2.getObligation() instanceof Obligation.Must);
		assertTrue(T.getObligation() instanceof Obligation.DisjunctiveMust);
		assertTrue(T1.getObligation() instanceof Obligation.Generic);
		assertTrue(T2.getObligation() instanceof Obligation.All);
		assertTrue(T3.getObligation() instanceof Obligation.DisjunctiveMust);

	}
}
